module dg1d

  using Reexport
  import Base: Filesystem, Threads, @propagate_inbounds
  using Configurations
  using DataFrames
  using DocStringExtensions
  using Glob
  using InteractiveUtils
  using Jacobi
  @reexport using HDF5
  using LinearAlgebra
  using Match
  import MacroTools
  using OrderedCollections
  using Plots
  @reexport import Parameters: @unpack, @with_kw
  using PrettyTables
  using Printf
  using ProgressMeter
  using REPL.TerminalMenus
  using StaticArrays
  using StructArrays
  @reexport using TOML

  export Maybe

  # utils
  include("utils.jl")
  include("numeric_utils.jl")
  include("objectcache.jl")
  include("rootfinders.jl")
  include("with_signature.jl")
  include("units.jl")
  include("parameters.jl")
  include("flatdict.jl")
  include("output.jl")
  include("refvector.jl")

  # differentiation and approximation
  include("lagrange.jl")
  include("gb.jl")
  include("glgl.jl")
  include("lgl.jl")
  include("spectralelement.jl")
  include("mesh.jl")

  # custom evolution
  include("evolve.jl")
  # evolution based on OrdinaryDiffEq pkg
  # using OrdinaryDiffEq
  # include("evolve_ordinarydiffeqs.jl")

  # interfaces
  include("equations.jl")
  include("variables.jl")

  include("tensorbasis.jl")
  include("box.jl")
  include("tree.jl")

  include("cache.jl")
  include("callbacks.jl")

  include("numericalflux.jl")
  include("riemannsolver.jl")
  include("bdryconditions.jl")
  include("dg_rhs.jl")

  include("segment.jl")

  ## projects
  include("projects/TCI/TCI.jl")
  @reexport using .TCI
  include("projects/HRSC/HRSC.jl")
  @reexport using .HRSC
  include("projects/EquationOfState/EquationOfState.jl")
  @reexport using .EquationOfState

  include("projects/ScalarEquation/ScalarEquation.jl")

  include("projects/EulerEq/EulerEq.jl")

  include("projects/SREulerEq/SREulerEq.jl")

  include("projects/TOV/TOV.jl")

  include("projects/SphericalAccretion/SphericalAccretion.jl")

  # include("projects/GRHD/GRHD.jl")

end # module
