using Test
using dg1d


# include("numeric_utils.jl")
# include("tree.jl")
# include("box.jl")


@testset "Cartesian coordinate charts" begin

  import dg1d: directions, orthogonal, flip

  @test directions(Cart1d.X) == (Cart1d.Xmin, Cart1d.Xmax)
  @test directions(Cart2d.X) == (Cart2d.Xmin, Cart2d.Xmax)
  @test directions(Cart2d.Y) == (Cart2d.Ymin, Cart2d.Ymax)
  @test directions(Cart3d.X) == (Cart3d.Xmin, Cart3d.Xmax)
  @test directions(Cart3d.Y) == (Cart3d.Ymin, Cart3d.Ymax)
  @test directions(Cart3d.Z) == (Cart3d.Zmin, Cart3d.Zmax)

  @test orthogonal(Cart1d.X) == ()
  @test orthogonal(Cart2d.X) == (Cart2d.Y,)
  @test orthogonal(Cart2d.Y) == (Cart2d.X,)
  @test orthogonal(Cart3d.X) == (Cart3d.Y, Cart3d.Z)
  @test orthogonal(Cart3d.Y) == (Cart3d.X, Cart3d.Z)
  @test orthogonal(Cart3d.Z) == (Cart3d.X, Cart3d.Y)

  @test flip(Cart1d.Xmin) == Cart1d.Xmax
  @test flip(Cart1d.Xmax) == Cart1d.Xmin
  @test flip(Cart2d.Xmin) == Cart2d.Xmax
  @test flip(Cart2d.Xmax) == Cart2d.Xmin
  @test flip(Cart2d.Ymin) == Cart2d.Ymax
  @test flip(Cart2d.Ymax) == Cart2d.Ymin
  @test flip(Cart3d.Xmin) == Cart3d.Xmax
  @test flip(Cart3d.Xmax) == Cart3d.Xmin
  @test flip(Cart3d.Ymin) == Cart3d.Ymax
  @test flip(Cart3d.Ymax) == Cart3d.Ymin
  @test flip(Cart3d.Zmin) == Cart3d.Zmax
  @test flip(Cart3d.Zmax) == Cart3d.Zmin

end


@testset "Cartesian coordinates" begin

  import dg1d: Coords1d, Coords2d, Coords3d,
         is_neighbor, touch_direction, distance, signed_distance

  # constructors require extract braces since Coords is just an alias of a NTuple
  Coords1d((1,))
  Coords2d((1,2))
  Coords3d((1,2,3))


  # is neighbor in any direction
  @test is_neighbor(Coords1d(1), Coords1d(2)) == true
  @test is_neighbor(Coords1d(1), Coords1d(3)) == false

  @test is_neighbor(Coords2d((1,1)), Coords2d((1,2))) == true
  @test is_neighbor(Coords2d((1,1)), Coords2d((2,1))) == true
  @test is_neighbor(Coords2d((1,1)), Coords2d((2,2))) == false

  @test is_neighbor(Coords3d((1,1,1)), Coords3d((1,2,1))) == true
  @test is_neighbor(Coords3d((1,1,1)), Coords3d((1,1,2))) == true
  @test is_neighbor(Coords3d((1,1,1)), Coords3d((2,1,1))) == true
  @test is_neighbor(Coords3d((1,1,1)), Coords3d((2,2,1))) == false
  @test is_neighbor(Coords3d((1,1,1)), Coords3d((1,2,2))) == false
  @test is_neighbor(Coords3d((1,1,1)), Coords3d((2,1,2))) == false
  @test is_neighbor(Coords3d((1,1,1)), Coords3d((2,2,2))) == false


  # is neighbor in specific direction
  @test is_neighbor(Coords1d(1), Coords1d(2), Cart1d.Xmax) == true
  @test is_neighbor(Coords1d(1), Coords1d(2), Cart1d.Xmin) == false
  @test is_neighbor(Coords1d(2), Coords1d(1), Cart1d.Xmax) == false
  @test is_neighbor(Coords1d(2), Coords1d(1), Cart1d.Xmin) == true
  @test is_neighbor(Coords1d(1), Coords1d(10), Cart1d.Xmin) == false
  @test is_neighbor(Coords1d(1), Coords1d(10), Cart1d.Xmax) == false


  @test is_neighbor(Coords2d((1,1)), Coords2d((2,1)), Cart2d.Xmax) == true
  @test is_neighbor(Coords2d((1,1)), Coords2d((2,1)), Cart2d.Xmin) == false
  @test is_neighbor(Coords2d((2,1)), Coords2d((1,1)), Cart2d.Xmax) == false
  @test is_neighbor(Coords2d((2,1)), Coords2d((1,1)), Cart2d.Xmin) == true
  @test is_neighbor(Coords2d((1,1)), Coords2d((1,2)), Cart2d.Xmax) == false
  @test is_neighbor(Coords2d((1,1)), Coords2d((1,2)), Cart2d.Xmin) == false

  @test is_neighbor(Coords2d((1,1)), Coords2d((1,2)), Cart2d.Ymax) == true
  @test is_neighbor(Coords2d((1,1)), Coords2d((1,2)), Cart2d.Ymin) == false
  @test is_neighbor(Coords2d((1,2)), Coords2d((1,1)), Cart2d.Ymax) == false
  @test is_neighbor(Coords2d((1,2)), Coords2d((1,1)), Cart2d.Ymin) == true
  @test is_neighbor(Coords2d((1,1)), Coords2d((2,1)), Cart2d.Ymax) == false
  @test is_neighbor(Coords2d((1,1)), Coords2d((2,1)), Cart2d.Ymin) == false

  @test is_neighbor(Coords2d((1,1)), Coords2d((5,5)), Cart2d.Xmin) == false
  @test is_neighbor(Coords2d((1,1)), Coords2d((5,5)), Cart2d.Xmax) == false
  @test is_neighbor(Coords2d((1,1)), Coords2d((5,5)), Cart2d.Ymin) == false
  @test is_neighbor(Coords2d((1,1)), Coords2d((5,5)), Cart2d.Ymax) == false


  @test is_neighbor(Coords3d((1,1,1)), Coords3d((2,1,1)), Cart3d.Xmax) == true
  @test is_neighbor(Coords3d((1,1,1)), Coords3d((2,1,1)), Cart3d.Xmin) == false
  @test is_neighbor(Coords3d((2,1,1)), Coords3d((1,1,1)), Cart3d.Xmax) == false
  @test is_neighbor(Coords3d((2,1,1)), Coords3d((1,1,1)), Cart3d.Xmin) == true
  @test is_neighbor(Coords3d((1,1,1)), Coords3d((1,2,1)), Cart3d.Xmin) == false
  @test is_neighbor(Coords3d((1,1,1)), Coords3d((1,2,1)), Cart3d.Xmax) == false
  @test is_neighbor(Coords3d((1,1,1)), Coords3d((1,1,2)), Cart3d.Xmin) == false
  @test is_neighbor(Coords3d((1,1,1)), Coords3d((1,1,2)), Cart3d.Xmax) == false

  @test is_neighbor(Coords3d((1,1,1)), Coords3d((1,2,1)), Cart3d.Ymax) == true
  @test is_neighbor(Coords3d((1,1,1)), Coords3d((1,2,1)), Cart3d.Ymin) == false
  @test is_neighbor(Coords3d((1,2,1)), Coords3d((1,1,1)), Cart3d.Ymax) == false
  @test is_neighbor(Coords3d((1,2,1)), Coords3d((1,1,1)), Cart3d.Ymin) == true
  @test is_neighbor(Coords3d((1,1,1)), Coords3d((2,1,1)), Cart3d.Ymin) == false
  @test is_neighbor(Coords3d((1,1,1)), Coords3d((2,1,1)), Cart3d.Ymax) == false
  @test is_neighbor(Coords3d((1,1,1)), Coords3d((1,1,2)), Cart3d.Ymin) == false
  @test is_neighbor(Coords3d((1,1,1)), Coords3d((1,1,2)), Cart3d.Ymax) == false

  @test is_neighbor(Coords3d((1,1,1)), Coords3d((1,1,2)), Cart3d.Zmax) == true
  @test is_neighbor(Coords3d((1,1,1)), Coords3d((1,1,2)), Cart3d.Zmin) == false
  @test is_neighbor(Coords3d((1,1,2)), Coords3d((1,1,1)), Cart3d.Zmax) == false
  @test is_neighbor(Coords3d((1,1,2)), Coords3d((1,1,1)), Cart3d.Zmin) == true
  @test is_neighbor(Coords3d((1,1,1)), Coords3d((2,1,1)), Cart3d.Zmin) == false
  @test is_neighbor(Coords3d((1,1,1)), Coords3d((2,1,1)), Cart3d.Zmax) == false
  @test is_neighbor(Coords3d((1,1,1)), Coords3d((1,2,1)), Cart3d.Zmin) == false
  @test is_neighbor(Coords3d((1,1,1)), Coords3d((1,2,1)), Cart3d.Zmax) == false

  @test is_neighbor(Coords3d((1,1,1)), Coords3d((5,5,5)), Cart3d.Xmin) == false
  @test is_neighbor(Coords3d((1,1,1)), Coords3d((5,5,5)), Cart3d.Xmax) == false
  @test is_neighbor(Coords3d((1,1,1)), Coords3d((5,5,5)), Cart3d.Ymin) == false
  @test is_neighbor(Coords3d((1,1,1)), Coords3d((5,5,5)), Cart3d.Ymax) == false
  @test is_neighbor(Coords3d((1,1,1)), Coords3d((5,5,5)), Cart3d.Zmin) == false
  @test is_neighbor(Coords3d((1,1,1)), Coords3d((5,5,5)), Cart3d.Zmax) == false


  @test touch_direction(Coords1d(1), Coords1d(1)) == nothing
  @test touch_direction(Coords1d(1), Coords1d(2)) == Cart1d.Xmax
  @test touch_direction(Coords1d(2), Coords1d(1)) == Cart1d.Xmin
  @test touch_direction(Coords2d((1,1)), Coords2d((1,1))) == nothing
  @test touch_direction(Coords2d((1,1)), Coords2d((2,1))) == Cart2d.Xmax
  @test touch_direction(Coords2d((2,1)), Coords2d((1,1))) == Cart2d.Xmin
  @test touch_direction(Coords2d((1,1)), Coords2d((1,2))) == Cart2d.Ymax
  @test touch_direction(Coords2d((1,2)), Coords2d((1,1))) == Cart2d.Ymin
  @test touch_direction(Coords3d((1,1,1)), Coords3d((1,1,1))) == nothing
  @test touch_direction(Coords3d((1,1,1)), Coords3d((2,1,1))) == Cart3d.Xmax
  @test touch_direction(Coords3d((2,1,1)), Coords3d((1,1,1))) == Cart3d.Xmin
  @test touch_direction(Coords3d((1,1,1)), Coords3d((1,2,1))) == Cart3d.Ymax
  @test touch_direction(Coords3d((1,2,1)), Coords3d((1,1,1))) == Cart3d.Ymin
  @test touch_direction(Coords3d((1,1,1)), Coords3d((1,1,2))) == Cart3d.Zmax
  @test touch_direction(Coords3d((1,1,2)), Coords3d((1,1,1))) == Cart3d.Zmin

end


@testset "Cell" begin

  import dg1d: is_neighbor

  c1d = Cell1d(neighbors=(1,2))
  c2d = Cell2d(neighbors=(1,2,3,0))
  c3d = Cell3d(neighbors=(1,2,3,4,5,0))

  @test is_boundarycell(c1d) == false
  @test is_boundarycell(c2d) == true
  @test is_boundarycell(c3d) == true

  c1, c2 = Cell1d(index=1), Cell1d(neighbors=(1,2))
  @test is_neighbor(c1, c2, Cart1d.Xmin) == true
  @test is_neighbor(c1, c2, Cart1d.Xmax) == false

  c1, c2 = Cell2d(index=2), Cell2d(neighbors=(1,2,3,4))
  @test is_neighbor(c1, c2, Cart2d.Xmin) == false
  @test is_neighbor(c1, c2, Cart2d.Xmax) == true
  @test is_neighbor(c1, c2, Cart2d.Ymin) == false
  @test is_neighbor(c1, c2, Cart2d.Ymax) == false

  c1, c2 = Cell3d(index=4), Cell3d(neighbors=(1,2,3,4,5,6))
  @test is_neighbor(c1, c2, Cart3d.Xmin) == false
  @test is_neighbor(c1, c2, Cart3d.Xmax) == false
  @test is_neighbor(c1, c2, Cart3d.Ymin) == false
  @test is_neighbor(c1, c2, Cart3d.Ymax) == true
  @test is_neighbor(c1, c2, Cart3d.Zmin) == false
  @test is_neighbor(c1, c2, Cart3d.Zmax) == false

end


@testset "Tree" begin

  Nx, Ny, Nz = 5, 6, 7

  # test with periodicity here, because that requires an extra step for setup
  t1d = Tree1d(Nx, periodic=(true,))
  t2d = Tree2d(Nx,Ny, periodic=(true,true))
  t3d = Tree3d(Nx,Ny,Nz, periodic=(true,true,true))

  @test n_cells(t1d) == Nx
  @test n_cells(t2d) == Nx*Ny
  @test n_cells(t3d) == Nx*Ny*Nz

  # t1d = Tree1d(5, periodic=(true,))
  # t2d = Tree2d(5,6, periodic=(true,false))
  # t3d = Tree3d(5,6,7, periodic=(true,false,true))

  # call show
  println(t1d)
  println(t2d)
  println(t3d)

  # iterator interface
  length_t1d = 0
  for c in t1d
    length_t1d += 1
  end
  @test length_t1d == n_cells(t1d)

  length_t2d = 0
  for c in t2d
    length_t2d += 1
  end
  @test length_t2d == n_cells(t2d)

  length_t3d = 0
  for c in t3d
    length_t3d += 1
  end
  @test length_t3d == n_cells(t3d)

  x_extends, y_extends, z_extends = (-2.0,2.0), (-10.0,0.0), (1.0,3.0)
  boxes1d = place_boxes(t1d, x_extends)
  boxes2d = place_boxes(t2d, x_extends, y_extends)
  boxes3d = place_boxes(t3d, x_extends, y_extends, z_extends)

  # # TODO Visualize generated trees to check for correctness
  # # TODO How to check if periodic boundaries are connected?
  # #      Perhaps also save is_boundarycell info and color boundary cells.
  # # for (name, tree) in [ ("tree1d", t1d), ("tree2d", t2d), ("tree3d", t3d) ]
  # #   open(filename, "w") do file
  # #     writedlm(file, tree.coords)
  # #   end
  # # end

end
